// Copyright (c) 2021, the Dart project authors.  Please see the AUTHORS file
// for details. All rights reserved. Use of this source code is governed by a
// BSD-style license that can be found in the LICENSE file.

part of alog;

const libyx_alog = 'libyx_alog.so';
/// 获取[DynamicLibrary]的runTime.
/// if so Failed to load dynamic library (dlopen failed: library \"libalog1
/// .so\" not found)
final DynamicLibrary dylib = Platform.isAndroid
    ? DynamicLibrary.open(libyx_alog)
    : DynamicLibrary.process();
/**
 * /// FFI signature of the hello_world C function
    typedef hello_world_func = ffi.Void Function();

    /// Dart type definition for calling the C foreign function
    typedef HelloWorld = void Function();

    /// Look up the C function 'hello_world'
    final HelloWorld hello = dylib
    .lookup<ffi.NativeFunction<hello_world_func>>('hello_world')
    .asFunction();
    // Call the function
    hello();
 */

/// FFI signature of the alogger_init C function
/// Map to the corresponding C interface
/// void alogger_init(int level, const char* path, const char* _name_prefix);
typedef alogger_init_native = Int32 Function(
    Int32, Pointer<Utf8>, Pointer<Utf8>);

/// Dart type definition for calling the C foreign function
typedef AloggerInit = int Function(int, Pointer<Utf8>, Pointer<Utf8>);

/// ALog 初始化
/// 使用 每次使用前调用 alogger_init 接口完成初始化，之后即可按目标日志等级调用 输出接口。
/// [AloggerInit] dart -> C [alogger_init_native],
///
///  Call the function : aloggerInit()
final aloggerInit =
    dylib.lookupFunction<alogger_init_native, AloggerInit>(alogger_init);

/// FFI signature of the alogger_init C function
///  [tag] tag 说明
///  [type] 0-普通log，1-api log
///  [moduleName] 模块名称
///  [line] 行数，不支持传 0 即可
///  [content] 日志内容
///
typedef alogger_func = Int32 Function(
    Pointer<Utf8>, Int32, Pointer<Utf8>, Int32, Int32, Pointer<Utf8>);

/// Dart type definition for calling the C foreign function
typedef AloggerDartFunc = int Function(
    Pointer<Utf8>, int, Pointer<Utf8>, int, int, Pointer<Utf8>);

/// ALog debug level
/// 使用 每次使用前调用 alogger_init 接口完成初始化，之后即可按目标日志等级调用 输出接口。
/// [AloggerDartFunc] dart -> C [alogger_func],
///
/// [fuc] C++ methmod name.
Function aloggerCallC(String fuc) =>
    dylib.lookupFunction<alogger_func, AloggerDartFunc>(fuc);

///Call dart function [aloggerCallC]
final aloggerVerbose = aloggerCallC(alogger_verbose);
final aloggerInfo = aloggerCallC(alogger_info);
final aloggerDebug = aloggerCallC(alogger_debug);
final aloggerTest = aloggerCallC(alogger_test);
final aloggerWarn = aloggerCallC(alogger_warn);
final aloggerError = aloggerCallC(alogger_error);

const alogger_init = 'alogger_init';
const alogger_verbose = 'alogger_verbose';
const alogger_info = 'alogger_info';
const alogger_debug = 'alogger_debug';
const alogger_test = 'alogger_test';
const alogger_warn = 'alogger_warn';
const alogger_error = 'alogger_error';

/// release
final int Function() aloggerRelease = dylib
    .lookup<NativeFunction<Int32 Function()>>("alogger_release")
    .asFunction();

/// aloggerPid
final int Function() aloggerPid =
    dylib.lookup<NativeFunction<Int32 Function()>>("alogger_pid").asFunction();

/// aloggerTid()
final int Function() aloggerTid =
    dylib.lookup<NativeFunction<Int32 Function()>>("alogger_tid").asFunction();

class AlogImpl implements AlogAbs {
  factory AlogImpl() => _getInstance();

  static AlogImpl get instance => _getInstance();

  static AlogImpl _instance = AlogImpl._internal();

  AlogImpl._internal();

  static AlogImpl _getInstance() => _instance;

  bool _isInit = false;

  /// C function: [aloggerInit]
  @override
  bool init(int level, String path, String namePrefix) {
    try {
      _isInit =
          aloggerInit(level, path.toNativeUtf8(), namePrefix.toNativeUtf8()) ==
              0;
    } catch (e) {
      _isInit = false;
      return _isInit;
    }

    /// 扫描对应C++是否存在对应function
    try {
      var tag = 'tag'.toNativeUtf8();
      var content = 'content'.toNativeUtf8();
      var moduleName = 'moduleName'.toNativeUtf8();
      aloggerVerbose(tag, 0, moduleName, 0, 0, content);
      aloggerInfo(tag, 0, moduleName, 0, 0, content);
      aloggerDebug(tag, 0, moduleName, 0, 0, content);
      aloggerWarn(tag, 0, moduleName, 0, 0, content);
      aloggerError(tag, 0, moduleName, 0, 0, content);
      aloggerTest(tag, 0, moduleName, 0, 0, content);
    } catch (e) {
      _isInit = false;
      return _isInit;
    }

    return _isInit;
  }

  @override
  void verbose(
      String tag, int type, String moduleName, int line, String content) {
    if (_isInit) {
      aloggerVerbose(tag.toNativeUtf8(), type, moduleName.toNativeUtf8(), line,
          aloggerTid(), content.toNativeUtf8());
    }
  }

  @override
  void info(String tag, int type, String moduleName, int line, String content) {
    if (_isInit) {
      aloggerInfo(tag.toNativeUtf8(), type, moduleName.toNativeUtf8(), line,
          aloggerTid(), content.toNativeUtf8());
    }
  }

  @override
  void warn(String tag, int type, String moduleName, int line, String content) {
    if (_isInit) {
      aloggerWarn(tag.toNativeUtf8(), type, moduleName.toNativeUtf8(), line,
          aloggerTid(), content.toNativeUtf8());
    }
  }

  @override
  void test(String tag, int type, String moduleName, int line, String content) {
    if (_isInit) {
      aloggerTest(tag.toNativeUtf8(), type, moduleName.toNativeUtf8(), line,
          aloggerTid(), content.toNativeUtf8());
    }
  }

  @override
  void debug(
      String tag, int type, String moduleName, int line, String content) {
    if (_isInit) {
      aloggerDebug(tag.toNativeUtf8(), type, moduleName.toNativeUtf8(), line,
          aloggerTid(), content.toNativeUtf8());
    }
  }

  @override
  void error(
      String tag, int type, String moduleName, int line, String content) {
    if (_isInit) {
      aloggerError(tag.toNativeUtf8(), type, moduleName.toNativeUtf8(), line,
          aloggerTid(), content.toNativeUtf8());
    }
  }

  @override
  int flushAsync() {
    // TODO: implement flushAsync
    throw UnimplementedError();
  }

  @override
  int flushSync() {
    // TODO: implement flushSync
    throw UnimplementedError();
  }

  @override
  int release() {
    return aloggerRelease();
  }

  @override
  int currentTid() {
    return aloggerTid();
  }
}
